__all__ = ["Itunes", "ItunesException"]

from datetime import datetime
from pprint import pprint
from typing import Optional

import requests

from yutipy.exceptions import InvalidValueException, ItunesException
from yutipy.logger import logger
from yutipy.lrclib import LrcLib
from yutipy.models import MusicInfo
from yutipy.utils.helpers import (
    are_strings_similar,
    guess_album_type,
    is_valid_string,
    separate_artists,
)


class Itunes:
    """A class to interact with the iTunes API."""

    def __init__(self, fetch_lyrics: bool = True) -> None:
        """
        Parameters
        ----------
        fetch_lyrics : bool, optional
            Whether to fetch lyrics (using `LRCLIB <https://lrclib.net>`__) if the music platform does not provide lyrics (default is True).
        """
        self.api_url = "https://itunes.apple.com"
        self.normalize_non_english = True
        self._is_session_closed = False
        self.fetch_lyrics = fetch_lyrics
        self.__session = requests.Session()
        self.__translation_session = requests.Session()

    def __enter__(self):
        """Enters the runtime context related to this object."""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        """Exits the runtime context related to this object."""
        self.close_session()

    def close_session(self) -> None:
        """Closes the current session."""
        if not self.is_session_closed:
            self.__session.close()
            self.__translation_session.close()
            self._is_session_closed = True

    @property
    def is_session_closed(self) -> bool:
        """Checks if the session is closed."""
        return self._is_session_closed

    def search(
        self,
        artist: str,
        song: str,
        limit: int = 10,
        normalize_non_english: bool = True,
    ) -> Optional[MusicInfo]:
        """
        Searches for a song by artist and title.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        limit: int, optional
            The number of items to retrieve from API. ``limit >=1 and <= 50``. Default is ``10``.
        normalize_non_english : bool, optional
            Whether to normalize non-English characters for comparison. Default is ``True``.

        Returns
        -------
        Optional[MusicInfo_]
            The music information if found, otherwise None.
        """
        if not is_valid_string(artist) or not is_valid_string(song):
            raise InvalidValueException(
                "Artist and song names must be valid strings and can't be empty."
            )

        self.normalize_non_english = normalize_non_english

        entities = ["song", "album"]
        for entity in entities:
            endpoint = f"{self.api_url}/search"
            query = f"?term={artist} - {song}&media=music&entity={entity}&limit={limit}"
            query_url = endpoint + query

            try:
                logger.info(
                    f'Searching iTunes for `artist="{artist}"` and `song="{song}"`'
                )
                logger.debug(f"Query URL: {query_url}")
                response = self.__session.get(query_url, timeout=30)
                logger.debug(f"Response status code: {response.status_code}")
                response.raise_for_status()
            except requests.RequestException as e:
                logger.warning(f"Unexpected error while searching iTunes: {e}")
                return None

            try:
                logger.debug("Parsing response JSON.")
                result = response.json()["results"]
            except (IndexError, KeyError, ValueError) as e:
                logger.warning(f"Invalid response structure from iTunes: {e}")
                return None

            music_info = self._parse_result(artist, song, result)
            if music_info:
                return music_info

        logger.warning(
            f"No matching results found for artist='{artist}' and song='{song}'"
        )
        return None

    def _parse_result(
        self, artist: str, song: str, results: list[dict]
    ) -> Optional[MusicInfo]:
        """
        Parses the search results to find a matching song.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        results : list
            The search results from the API.

        Returns
        -------
        Optional[MusicInfo]
            The music information if a match is found, otherwise None.
        """
        for result in results:
            if not (
                are_strings_similar(
                    result.get("trackName", result["collectionName"]),
                    song,
                    use_translation=self.normalize_non_english,
                    translation_session=self.__translation_session,
                )
                and are_strings_similar(
                    result["artistName"],
                    artist,
                    use_translation=self.normalize_non_english,
                    translation_session=self.__translation_session,
                )
            ):
                continue

            album_title, album_type = self._extract_album_info(result)
            release_date = self._format_release_date(result["releaseDate"])
            artists = separate_artists(result["artistName"])

            music_info = MusicInfo(
                album_art=result["artworkUrl100"],
                album_title=album_title,
                album_type=album_type.lower(),
                artists=", ".join(artists),
                genre=result["primaryGenreName"],
                id=result.get("trackId", result["collectionId"]),
                isrc=None,
                lyrics=None,
                release_date=release_date,
                tempo=None,
                title=result.get("trackName", album_title),
                type=result["wrapperType"],
                upc=None,
                url=result.get("trackViewUrl", result["collectionViewUrl"]),
            )

            if self.fetch_lyrics:
                with LrcLib() as lrc_lib:
                    lyrics = lrc_lib.get_lyrics(
                        artist=music_info.artists, song=music_info.title
                    )
                if lyrics:
                    music_info.lyrics = lyrics.get("plainLyrics")

            return music_info
        return None

    def _extract_album_info(self, result: dict) -> tuple:
        """
        Extracts album information from a search result.

        Parameters
        ----------
        result : dict
            A single search result from the API.

        Returns
        -------
        tuple
            The extracted album title and type.
        """

        guess = guess_album_type(result.get("trackCount", 1))
        guessed_right = are_strings_similar(
            result.get("wrapperType", "x"), guess, use_translation=False
        )
        return result["collectionName"], (
            result["wrapperType"] if guessed_right else guess
        )

    def _format_release_date(self, release_date: str) -> str:
        """
        Formats the release date to a standard format.

        Parameters
        ----------
        release_date : str
            The release date from the API.

        Returns
        -------
        str
            The formatted release date.
        """
        return datetime.strptime(release_date, "%Y-%m-%dT%H:%M:%SZ").strftime(
            "%Y-%m-%d"
        )


if __name__ == "__main__":
    import logging

    from yutipy.logger import enable_logging

    enable_logging(level=logging.DEBUG)
    itunes = Itunes()

    try:
        artist_name = input("Artist Name: ")
        song_name = input("Song Name: ")
        pprint(itunes.search(artist_name, song_name))
    finally:
        itunes.close_session()



import pytest

from yutipy.listenbrainz import ListenBrainz
from tests import BaseResponse


@pytest.fixture
def listenbrainz():
    return ListenBrainz()


class MockResponseUsers(BaseResponse):
    @staticmethod
    def json():
        return {
            "users": [
                {"user_name": "test4test"},
                {"user_name": "lb_test"},
                {"user_name": "lb-test"},
                {"user_name": "muz-test"},
                {"user_name": "tes--"},
                {"user_name": "lb_test_1"},
                {"user_name": "suvid_test"},
                {"user_name": "testins"},
                {"user_name": "hemang-test"},
                {"user_name": "Test4585"},
            ]
        }


class MockResponseActivity(BaseResponse):
    @staticmethod
    def json():
        return {
            "payload": {
                "count": 1,
                "listens": [
                    {
                        "playing_now": True,
                        "track_metadata": {
                            "additional_info": {
                                "duration": 269,
                                "media_player": "Some Music Player",
                                "origin_url": "https://music.youtube.com/watch?v=lYBUbBu4W08",
                                "submission_client": "Some Scrobbler Plugin",
                            },
                            "artist_name": "Artist",
                            "track_name": "Song",
                        },
                    }
                ],
                "playing_now": True,
                "user_id": "Username",
            }
        }


@pytest.fixture
def mock_response_users(listenbrainz, monkeypatch):
    def mock_get(*args, **kwargs):
        return MockResponseUsers()

    monkeypatch.setattr(listenbrainz._ListenBrainz__session, "get", mock_get)


@pytest.fixture
def mock_response_activity(listenbrainz, monkeypatch):
    def mock_get(*args, **kwargs):
        return MockResponseActivity()

    monkeypatch.setattr(listenbrainz._ListenBrainz__session, "get", mock_get)


def test_find_user(listenbrainz, mock_response_users):
    username = "Test4585"
    user_name = listenbrainz.find_user(username=username)
    assert user_name is not None
    assert username == user_name

    username = "potato"
    user_name = listenbrainz.find_user(username=username)
    assert user_name is None


def test_find_user_case_insensitive(listenbrainz, mock_response_users):
    username = "muz-TEST"
    user_name = listenbrainz.find_user(username=username)
    assert user_name is not None
    assert username.lower() == user_name.lower()


def test_get_currently_playing(listenbrainz, mock_response_activity):
    username = "Username"
    activity = listenbrainz.get_currently_playing(username)
    assert activity.is_playing is True
    assert activity.title
    assert activity.artists
    assert activity.url

def test_get_currently_playing_no_user(listenbrainz, mock_response_activity):
    username = "Bob"
    activity = listenbrainz.get_currently_playing(username)
    assert activity is None



__all__ = ["ListenBrainz"]

from typing import Optional
from time import time

import requests

from yutipy.logger import logger
from yutipy.models import UserPlaying
from yutipy.utils.helpers import are_strings_similar


class ListenBrainz:
    """A class to interact with the ListenBrainz API for fetching user music data."""

    def __init__(self):
        self._is_session_closed = False
        self.__api_url = "https://api.listenbrainz.org"
        self.__session = requests.Session()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close_session()

    def close_session(self):
        """Closes the current session(s)."""
        if not self._is_session_closed:
            self.__session.close()
            self._is_session_closed = True

    @property
    def is_session_closed(self) -> bool:
        """Checks if the session is closed."""
        return self._is_session_closed

    def find_user(self, username: str) -> Optional[str]:
        """
        Whether profile with the provided username exists on the ListenBrainz.

        It searches ListenBrainz for the provided username and fuzzy matches
        the provided username with the username(s) returned from ListenBrainz API.

        Parameters
        ----------
        username : str
            The username to search.

        Returns
        -------
        Optional[str]
            The username returned by ListenBrainz API if user found,
            or ``None`` if the user with provided username does not exist.
        """
        endpoint = "/1/search/users/"
        url = self.__api_url + endpoint

        try:
            response = self.__session.get(
                url=url, params={"search_term": username}, timeout=30
            )
            response.raise_for_status()
        except requests.RequestException as e:
            logger.warning(f"Failed to search for the user: {e}")
            return

        response_json = response.json()
        users = response_json.get("users")

        if not users:
            return

        for user in users:
            user_name = user.get("user_name")
            if are_strings_similar(
                username, user_name, threshold=100, use_translation=False
            ):
                return user_name

    def get_currently_playing(self, username: str) -> Optional[UserPlaying]:
        """
        Fetches information about the currently playing track for a user.

        Parameters
        ----------
        username : str
            The ListenBrainz username to fetch data for.

        Returns
        -------
        Optional[UserPlaying_]
            An instance of the ``UserPlaying`` model containing details about the currently
            playing track if available, or ``None`` if the request fails or no data is available.
        """
        endpoint = f"/1/user/{username}/playing-now"
        url = self.__api_url + endpoint

        try:
            response = self.__session.get(url=url, timeout=30)
            response.raise_for_status()
        except requests.RequestException as e:
            logger.warning(f"Failed to retrieve listening activity: {e}")
            return

        response_json = response.json()
        if not response_json:
            logger.info(
                f"It seems no activity found for `{username}`. Might be user does not exist or not no data available."
            )
            return

        result = response_json.get("payload", {})
        is_playing = result.get("playing_now", False)
        user_id = result.get("user_id", "")

        if username.lower() != user_id.lower():
            return

        if result and is_playing:
            track_metadata = result.get("listens", [{}])[0].get("track_metadata", {})
            return UserPlaying(
                artists=track_metadata.get("artist_name"),
                id=None,
                timestamp=time(),
                title=track_metadata.get("track_name"),
                url=track_metadata.get("additional_info", {}).get("origin_url"),
                is_playing=is_playing,
            )
        return None


if __name__ == "__main__":
    with ListenBrainz() as listenbrainz:
        username = input("Enter ListenBrainz Username: ").strip()
        result = listenbrainz.find_user(username)

        if result:
            print(f"User found with username: {result}")
        else:
            print(f"No user with username '{username}' found!")



import pytest

from yutipy.lastfm import LastFm
from yutipy.models import UserPlaying
from tests import BaseResponse


@pytest.fixture
def lastfm():
    return LastFm(api_key="test_api_key")


class MockResponseActivity(BaseResponse):
    @staticmethod
    def json():
        return {
            "recenttracks": {
                "track": [
                    {
                        "artist": {"mbid": "", "#text": "Test Artist"},
                        "image": [
                            {
                                "size": "small",
                                "#text": "https://example.com/image/small.jpg",
                            },
                            {
                                "size": "extralarge",
                                "#text": "https://example.com/image/extralarge.jpg",
                            },
                        ],
                        "mbid": "",
                        "album": {
                            "mbid": "",
                            "#text": "Test Album",
                        },
                        "name": "Test Track",
                        "@attr": {"nowplaying": "true"},
                        "url": "https://www.last.fm/music/test+track",
                    }
                ]
            }
        }


class MockResponseProfile(BaseResponse):
    @staticmethod
    def json():
        return {
            "user": {
                "name": "john",
                "realname": "Real John",
                "image": [
                    {
                        "size": "small",
                        "#text": "https://example.com/image/john",
                    },
                    {
                        "size": "extralarge",
                        "#text": "https://example.com/image/john",
                    },
                ],
                "url": "https://example.com/john",
                "type": "user",
            }
        }


@pytest.fixture
def mock_response_activity(lastfm, monkeypatch):
    def mock_get(*args, **kwargs):
        return MockResponseActivity()

    monkeypatch.setattr(lastfm._LastFm__session, "get", mock_get)


@pytest.fixture
def mock_response_profile(lastfm, monkeypatch):
    def mock_get(*args, **kwargs):
        return MockResponseProfile()

    monkeypatch.setattr(lastfm._LastFm__session, "get", mock_get)


def test_get_currently_playing(lastfm, mock_response_activity):
    username = "bob"
    currently_playing = lastfm.get_currently_playing(username=username)
    assert currently_playing is not None
    assert isinstance(currently_playing, UserPlaying)
    assert currently_playing.title == "Test Track"
    assert currently_playing.album_title == "Test Album"
    assert "extralarge" in currently_playing.album_art
    assert currently_playing.is_playing is True


def test_get_user_profile(lastfm, mock_response_profile):
    username = "john"
    profile = lastfm.get_user_profile(username=username)
    assert profile is not None
    assert profile["username"] == username
    assert profile["name"] == "Real John"
    assert profile["type"] == "user"


def test_invalid_username(lastfm, monkeypatch):
    def mock_get(*args, **kwargs):
        class MockResponse(BaseResponse):
            @staticmethod
            def json():
                return {"message": "User not found", "error": 6}

        return MockResponse()

    monkeypatch.setattr(lastfm._LastFm__session, "get", mock_get)

    username = "alksdjfalsjdfweurppqoweiuwu"
    profile = lastfm.get_user_profile(username=username)
    assert profile is not None
    assert "error" in profile



import pytest
from pytest import raises

from yutipy.exceptions import InvalidValueException
from yutipy.models import MusicInfo
from yutipy.musicyt import MusicYT


@pytest.fixture
def music_yt():
    return MusicYT()


@pytest.fixture
def mock_ytmusic_search(music_yt, monkeypatch):
    def mock_search(*args, **kwargs):
        return [
            {
                "category": "Songs",
                "resultType": "song",
                "videoId": "ZrOKjDZOtkA",
                "title": "Test Song",
                "artists": [{"name": "Test Artist", "id": "ADlkasoiuUUer34ldb"}],
                "album": {"name": "Test Album", "id": "MIekd34_934"},
                "microformat": {
                    "microformatDataRenderer": {"uploadDate": "1969-12-31"}
                },
                "thumbnails": [{"url": "https://example.com/image/ZrOKjDZOtkA"}],
            }
        ]

    def mock_get_watch_playlist(*args, **kwargs):
        return {"lyrics": "MPLYt_HNNclO0Ddoc-17"}

    def mock_get_lyrics(*args, **kwargs):
        return {"lyrics": "Never Gonna Give You Up!"}

    # Patch the `search` and other methods of the `YTMusic` class
    monkeypatch.setattr("ytmusicapi.YTMusic.search", mock_search)
    monkeypatch.setattr(
        "ytmusicapi.YTMusic.get_watch_playlist", mock_get_watch_playlist
    )
    monkeypatch.setattr("ytmusicapi.YTMusic.get_lyrics", mock_get_lyrics)


def test_search_valid(music_yt, mock_ytmusic_search):
    artist = "Test Artist"
    song = "Test Song"
    result = music_yt.search(artist, song, normalize_non_english=False)
    assert result is not None
    assert isinstance(result, MusicInfo)
    assert artist in result.artists
    assert result.title == song


def test_search_invalid(music_yt, mock_ytmusic_search):
    artist = ";laksjdflkajsdfj;asdjf"
    song = "jaksjd;fljkas;dfkjasldkjf"
    result = music_yt.search(artist, song)
    assert result is None


def test_search_empty_artist(music_yt, mock_ytmusic_search):
    artist = ""
    song = "Song"

    with raises(InvalidValueException):
        music_yt.search(artist, song, normalize_non_english=False)


def test_search_empty_song(music_yt, mock_ytmusic_search):
    artist = "Artist"
    song = ""

    with raises(InvalidValueException):
        music_yt.search(artist, song, normalize_non_english=False)


def test_close_session(music_yt):
    music_yt.close_session()
    assert music_yt.is_session_closed



__all__ = ["LrcLib"]

from importlib.metadata import PackageNotFoundError, version
from typing import Optional

import requests

from yutipy.exceptions import InvalidValueException
from yutipy.logger import logger
from yutipy.utils import are_strings_similar, is_valid_string


class LrcLib:
    """
    A class to interact with the `LRCLIB <lrclib.net>`_ API for fetching lyrics.
    """

    def __init__(
        self,
        app_name: str = "yutipy",
        app_version: str = None,
        app_url: str = "https://github.com/CheapNightbot/yutipy",
    ) -> None:
        """
        Parameters
        ----------
        app_name : str
            The name of the application.
        app_version : str, optional
            The version of the application.
        app_url : str, optional
            The URL of the application.

        Notes
        -----
        These are used to set the User-Agent header for requests made to the API as suggested by the API documentation of `LRCLIB <lrclib.net>`_.
        """
        self.api_url = "https://lrclib.net/api"
        self.app_name = app_name
        self.app_url = app_url
        if not app_version:
            try:
                self.app_version = f"v{version('yutipy')}"
            except PackageNotFoundError:
                self.app_version = "N/A"
        else:
            self.app_version = app_version

        self._is_session_closed = False
        self.__session = requests.Session()
        self.__session.headers.update(
            {"User-Agent": f"{self.app_name} {self.app_version} ({self.app_url})"}
        )
        self._translation_session = requests.Session()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close_session()

    def close_session(self):
        """
        Closes the session if it is not already closed.
        """
        if not self._is_session_closed:
            self.__session.close()
            self._translation_session.close()
            self._is_session_closed = True

    @property
    def is_session_closed(self) -> bool:
        return self._is_session_closed

    def get_lyrics(
        self,
        artist: str,
        song: str,
        album: str = None,
        normalize_non_english: bool = True,
    ) -> Optional[dict]:
        """
        Fetches lyrics for a given artist and song.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        album : str, optional
            The title of the album.
        normalize_non_english : bool, optional
            Whether to normalize non-English characters for comparison (default is True).

        Returns
        -------
        Optional[dict]
            The lyrics information if found, otherwise None.
        """

        if not is_valid_string(artist) or not is_valid_string(song):
            raise InvalidValueException(
                "Artist and song names must be valid strings and can't be empty."
            )

        endpoint = f"{self.api_url}/search"
        query = f"?artist_name={artist}&track_name={song}"
        query += f"&album_name={album}" if album else ""
        query_url = endpoint + query

        try:
            logger.info(
                f"Fetching lyrics for artist: {artist}, song: {song}, album: {album}"
            )
            response = self.__session.get(query_url, timeout=30)
            logger.debug(f"Response status code: {response.status_code}")
            response.raise_for_status()
        except requests.RequestException as e:
            logger.warning(f"Unexpected error while fetching lyrics: {e}")
            return

        results = response.json()

        for result in results:
            if are_strings_similar(
                result.get("trackName"),
                song,
                use_translation=normalize_non_english,
                translation_session=self._translation_session,
            ) and are_strings_similar(
                result.get("artistName"),
                artist,
                use_translation=normalize_non_english,
                translation_session=self._translation_session,
            ):
                if album and not are_strings_similar(
                    result.get("albumName"),
                    album,
                    use_translation=normalize_non_english,
                    translation_session=self._translation_session,
                ):
                    continue
                return result
        return None


if __name__ == "__main__":
    import logging

    from yutipy.logger import enable_logging

    enable_logging(level=logging.DEBUG)

    with LrcLib() as lyric_lib:
        artist_name = input("Artist Name: ").strip()
        song_title = input("Song Title: ").strip()
        lyrics = lyric_lib.get_lyrics(artist_name, song_title)
        print(f"\nLyrics for '{song_title}' by {artist_name}:\n{'-' * 40}\n")

        if lyrics:
            print(lyrics.get("plainLyrics"))
        else:
            print(
                "It seems that the lyrics were not found! You might have to guess them..."
            )
